-- #####################################################################################
-- ## Skript zum Vortrag <die schwierigen Kopfnsse mit T-SQL gelst>                 ##     
-- #####################################################################################
-- ## Dieses Skript legt die Funktion fr die Aufgabe LEVENSHTEIN an. Es wird durch   ##
-- ## weitere Skripte ergnzt, die in ihrer durchnummerierten Reihenfolge abzu-       ##
-- ## arbeiten sind. Die Skripte passen zu den Ausfhrungen aus der zugehrigen       ##
-- ## Powerpoint-Prsentation vom selben Autor.                                       ##
-- ## ------------------------------------------------------------------------------- ##
-- ## Bitte dieses Skript niemals in einer Produktionsumgebung einsetzen, da          ##
-- ## Datenbankobjekte angelegt, ergnzt, gendert oder sogar gelscht werden!!!      ##
-- ## ------------------------------------------------------------------------------- ##
-- ## Verwendung auf eigene Gefahr...                                                 ##
-- ## ------------------------------------------------------------------------------- ##
-- ## Torsten Ahlemeyer, arelium GmbH        torsten.ahlemeyer@arelium.de             ##
-- #####################################################################################

-- auf die Projektdatenbank wechseln
USE [Workshop_Kopfnuesse]
GO

SET ANSI_NULLS ON
GO

SET QUOTED_IDENTIFIER ON
GO

IF NOT EXISTS ( SELECT  * FROM    sys.schemas  WHERE   name = N'Levenshtein' ) 

    EXEC('CREATE SCHEMA [Levenshtein]');
GO


-- =============================================
-- Berechnet die Levenshtein-Distanz zwischen zwei gegebenen Strings.
-- Gibt also zurck, wieviele atamore (1 Zeichen!) Ersetze-, Einfge- oder Lschschritte
-- notwendig sind, um einen String in einen anderen zu transformierten.
--
-- Vergleichoperationen verwenden die auf SQL-Server-Ebene konfigurierte Sensitivitt 
-- (standardmig wird die Gro-/Kleinschreibung nicht beachtet).
--
-- Basierend auf Sten Hjelmqvists "Fast, Memory Efficiency" Algorithmus, beschrieben
-- unter http://www.codeproject.com/Articles/13525/Fast-memory-efficient-Levenshtein-algorithm,
-- mit einigen zustzlichen Optimierungen.
-- =============================================

CREATE OR ALTER FUNCTION [Levenshtein].[Levenshtein](
    @Begriff1 NVARCHAR(4000)                                               -- erster bergebener Begriff
  , @Begriff2 NVARCHAR(4000)                                               -- zweiter bergebener Begriff
  , @Maximum INTEGER                                                       -- maximal erlaubter Unterschied, bevor NULL zurckgegeben wird 
	)
RETURNS INTEGER
WITH SCHEMABINDING
AS
BEGIN
	DECLARE @Distanz					INTEGER			= 0		-- return variable
	DECLARE @Zwischenregister			NVARCHAR(4000)			-- Laufendes Zwischenregister zum Speichern berechneter Distanzen
	DECLARE @Start						INTEGER			= 1		-- Index (Erstwert ist 1) des ersten abweichenden Zeichen
																--   zwischen den zwei bergebenen Begriffen
	DECLARE @SchleifeBegriff1			INTEGER					-- Schleifenzhler fr den ersten Begriff
	DECLARE @SchleifeBegriff2			INTEGER					-- Schleifenzhler fr den zweiten Begriff
	DECLARE @Diagonaldistanz			INTEGER					-- Distanz der Zellen nach links oben innerhalb einer "n x m"-Matrix
	DECLARE @Links						INTEGER					-- Distanz der Zellen nach links innerhalb einer "n x m"-Matrix
	DECLARE @ZeichenBegriff1			NCHAR(1)				-- Zeichen aus Begiff1
	DECLARE @AktzuellesZeichenBegriff2  INTEGER					-- Zwischenspeicherung von @SchleifeBegriff2 um sauber SELECTieren zu knnen
	DECLARE @SchleifeBegriff2Offset		INTEGER					-- Abstand zur Berechnung des Anfangswertes fr @SchleifeBegriff2
	DECLARE @SchleifeBegriff2End		INTEGER					-- Endwert fr @SchleifeBegriff2

	-- Berechnung der Lnge der Begriffe einschlielich aller abschlieenden Leerzeichen (die SQL Server sonst ignorieren wrde)
	DECLARE @Begriff1Laenge             INTEGER			= datalength(@Begriff1) / datalength(left(left(@Begriff1, 1) + '.', 1))    
																-- Lnge des ersten und hoffentlich krzeren(!) Begriffs
	DECLARE @Begriff2Laenge				INTEGER			= datalength(@Begriff2) / datalength(left(left(@Begriff2, 1) + '.', 1))    
																-- Lnge des zweiten und hoffentlich lngeren(!) Begriffs
	DECLARE @Laengendifferenz			INTEGER					-- Differnez der gerade berechneten Begriffslngen

	-- Wenn die Begriffe eine unterschiedliche Lnge aufweisen, ist es klug sicher zu stellen, dass der erste Begriff der krzere ist.
	-- Sollte dies noch nicht der Fall sein, werden die Begriffe hier vertauscht. Der Algorithmus geht davon aus, dass die zu absolvierenden 
	-- Schleifendurchlufe diesen geschwindigkeitsoptimierten Zustand vorfinden.
	IF (@Begriff1Laenge > @Begriff2Laenge) 
	BEGIN
		SELECT    @Zwischenregister	= @Begriff1
				, @SchleifeBegriff1	= @Begriff1Laenge				-- temopre Auslagerung eines Begriffsinhaltes 
		SELECT	  @Begriff1			= @Begriff2						-- ins Zwischenregister
				, @Begriff1Laenge   = @Begriff2Laenge
		SELECT	  @Begriff2			= @Zwischenregister
				, @Begriff2Laenge   = @SchleifeBegriff1
	END

	-- Sollten die beiden Begriffe schon anhand ihrer Lnge zu stark voneinander abweichen, wird die Funktion "abgebrochen"
    SELECT		  @Maximum			= ISNULL(@Maximum, @Begriff2Laenge)
				, @Laengendifferenz	= @Begriff2Laenge - @Begriff1Laenge
    IF @Laengendifferenz > @Maximum RETURN NULL

	-- Ein identisches Ende beider Begriffe kann man problemlos "rausrechnen" - die Begriffe werden entsprechend (virtuell) gekrzt, das
	-- spart Laufzeit fr den Restalgorithmus
    WHILE(@Begriff1Laenge > 0 AND SUBSTRING(@Begriff1, @Begriff1Laenge, 1) = SUBSTRING(@Begriff2, @Begriff2Laenge, 1))
        SELECT @Begriff1Laenge = @Begriff1Laenge - 1, @Begriff2Laenge = @Begriff2Laenge - 1

    -- Sollte der erste Begriff leer sein, wird die Lnge des zweiten Begriffes als Ergebnis genutzt (nur Inserts)
    IF (@Begriff1Laenge = 0) RETURN @Begriff2Laenge
    
	-- Einen identischen Anfang beider Begriffe kann man problemlos "rausrechnen" - die Begriffe werden entsprechend (virtuell) gekrzt, das
	-- spart Laufzeit fr den Restalgorithmus
    WHILE (@Start < @Begriff1Laenge AND SUBSTRING(@Begriff1, @Start, 1) = SUBSTRING(@Begriff2, @Start, 1)) 
        SELECT @Start = @Start + 1

    IF (@Start > 1) 
	BEGIN
		SELECT @Begriff1Laenge = @Begriff1Laenge - (@Start - 1)
				, @Begriff2Laenge = @Begriff2Laenge - (@Start - 1)

		-- wenn alle krzeren Strings mit Prfix und/oder Suffix von lngeren Strings bereinstimmen, dann ist die 
		-- Levenshtein-Distanz nur das Lschen zustzlicher Zeichen, die in lngeren Strings vorhanden sind
		IF (@Begriff1Laenge <= 0) RETURN @Begriff2Laenge

		SELECT	  @Begriff1 = SUBSTRING(@Begriff1, @Start, @Begriff1Laenge)
				, @Begriff2 = SUBSTRING(@Begriff2, @Start, @Begriff2Laenge)
	END

	-- Die Vereinfachungen sind abgeschlossen, nun muss der eigentliche Algorithmus loslegen...
	-- Es wird ein Array von Distanzen initialisiert
    SELECT    @Zwischenregister			=   ''
			, @SchleifeBegriff2			=	1

    WHILE (@SchleifeBegriff2            <=  @Begriff2Laenge) 
	BEGIN
		SELECT @Zwischenregister		=   @Zwischenregister + NCHAR(CASE WHEN @SchleifeBegriff2 > @Maximum THEN @Maximum ELSE @SchleifeBegriff2 END)
		SELECT @SchleifeBegriff2		=   @SchleifeBegriff2 + 1
    END

    SELECT    @SchleifeBegriff2Offset	=   @Maximum - @Laengendifferenz
			, @SchleifeBegriff1			=	1

    WHILE (@SchleifeBegriff1            <=	@Begriff1Laenge) 
    BEGIN
        SELECT @Distanz					=	@SchleifeBegriff1
             , @Diagonaldistanz         =	@SchleifeBegriff1 - 1
             , @ZeichenBegriff1         =	SUBSTRING(@Begriff1, @SchleifeBegriff1, 1)

             -- Nicht zu betrachten ist die Diagonale nach oben links: (@SchleifeBegriff1) + @Maximum
             -- Nicht zu betrachten ist die Diagonale nach unten rechts: (@SchleifeBegriff1 - @Laengendifferenz) - @Maximum 
             , @SchleifeBegriff2		=	CASE
												WHEN @SchleifeBegriff1 <= @SchleifeBegriff2Offset 
												THEN 1 
												ELSE @SchleifeBegriff1 - @SchleifeBegriff2Offset 
											END
             , @SchleifeBegriff2End		=	CASE 
                                                WHEN @SchleifeBegriff1 + @Maximum >= @Begriff2Laenge 
                                                THEN @Begriff2Laenge 
                                                ELSE @SchleifeBegriff1 + @Maximum 
											END

        WHILE (@SchleifeBegriff2 <= @SchleifeBegriff2End) 
		BEGIN
            -- An diesem Punkt enthlt @Distanz den vorherigen Wert (die Zelle oben, wenn wir eine m mal n-Matrix verwenden)
            SELECT @Links						= UNICODE(SUBSTRING(@Zwischenregister, @SchleifeBegriff2, 1))
                 , @AktzuellesZeichenBegriff2	= @SchleifeBegriff2

            SELECT @Distanz				= 
											CASE
												WHEN (@ZeichenBegriff1 = SUBSTRING(@Begriff2, @SchleifeBegriff2, 1)) 
												THEN @Diagonaldistanz													-- Match, kein Hochzhlen der Distanz
											ELSE 1 + 
												CASE 
													WHEN @Diagonaldistanz < @Links AND @Diagonaldistanz < @Distanz 
                                                    THEN @Diagonaldistanz												-- Austausch eines einzelnen Zeichens
													WHEN @Links < @Distanz 
                                                    THEN @Links															-- Einfgung
                                                    ELSE @Distanz														-- Lschung
												END
											END

            SELECT @Zwischenregister    = STUFF(@Zwischenregister, @AktzuellesZeichenBegriff2, 1, NCHAR(@Distanz))
                 , @Diagonaldistanz		= @Links
                 , @SchleifeBegriff2    =
											CASE 
												WHEN (@Distanz > @Maximum) AND (@AktzuellesZeichenBegriff2 = @SchleifeBegriff1 + @Laengendifferenz) 
												THEN @SchleifeBegriff2End + 2 
												ELSE @AktzuellesZeichenBegriff2 + 1 
											END
        END

        SELECT @SchleifeBegriff1		= 
											CASE 
												WHEN @SchleifeBegriff2 > @SchleifeBegriff2End + 1 
														THEN @Begriff1Laenge + 1 
												ELSE @SchleifeBegriff1 + 1 
											END
    END
    RETURN CASE 
        WHEN @Distanz <= @Maximum 
                THEN @Distanz 
        ELSE NULL 
    END
END
